/*****************************************************************************
MODULE: minmax.cpp
PROJECT: minmax
AUTHOR: Jeff Nielsen
DATE: January 1996
DESCRIPTION: An implementation of the lower-bound algorithm for finding the
             minimum & maximum of a set of numbers.  Annotated with SAIL
	     function calls for visualizing the process.

*****************************************************************************/
#include <iostream.h>
#include <strstrea.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "stdjeff.h"
#define SWAN

#define MAXLENGTH 15		   // Maximum size of buckets
#define MAXINT 1000		       // Maximum number generated

// SWAN stuff -------------------------------------------------------------
#ifdef SWAN
#include "sail.h"
char memstr[100];	      // Global memory space for string formatting
ostrstream memstm(memstr, 100);
GRAPHID topg, lessg, moreg, finalg,    // Swan graph names
        countg = GRAPHID(100); 

//////////////////////////////////////////////////
inline void make_swangraph(GRAPHID gid, NODETYPE nt, bool dir_undir,
            LAYOUTCOMPTYPE lctype)
//
// PURPOSE: Macro for making SWAN graphs
// PARAMETERS:
//    gid: The graphid you want for this graph
//    nt: The type of node you want the graph to have
//    dir_undir: DIRECTED or UNDIRECTED graph
//    lctype: The default layout component you want this graph to have.
{
   sw_newgraph(gid, nt, dir_undir);
   sw_newlc(gid, LC_ID(1), lctype);
   sw_setcurlc(gid, LC_ID(1));
}

#endif


// Bucket class definition ---------------------------------------------------
class Bucket {
private:
   int *bkt;			       // Pointer to array
   int num_vals, max_vals;
   static int compcounter;	       // Number of comparisons

public:
   Bucket(int size) { bkt = new int[size];  num_vals = 0; max_vals = size; }
   ~Bucket() { delete[] bkt; }

   inline void add(int);		// Adds element to end of bucket
   int remove();			// Removes first element of bucket
   int getelem(int num) { return bkt[num]; }
   void print();

   int count() { return num_vals; };
   bool is_empty() { return num_vals == 0; }

   inline bool smaller();	// Compares first two elements in bucket
   void swap()			// Swap first two elements in bucket
      { int save = bkt[0]; bkt[0] = bkt[1]; bkt[1] = save; }
   int comparison_count() { return compcounter; }
};

//////////////////////////////////////////////////
void Bucket::add(int val)
//
// PURPOSE: Adds a value to the bucket
// PARAMETER:
//    val: Value to add
{
   bkt[num_vals++] = val; 
#ifdef SWAN
   memstm.seekp(0);
   memstm << val << ends;
   sw_setnodeattr(GRAPHID(this), NODEID(num_vals-1), GNODELABEL, memstr);
#endif
}

//////////////////////////////////////////////////
int Bucket::remove()
//
// PURPOSE: Removes the first value from the bucket
// RETURNS: The value removed
{
   int i, save = bkt[0];
   for (i = 1; i < num_vals; i++) {	  // Move all other values up
      bkt[i-1] = bkt[i];
#ifdef SWAN
      memstm.seekp(0);
      memstm << getelem(i-1) << ends;
      sw_setnodeattr(GRAPHID(this), NODEID(i-1), GNODELABEL, memstr);
#endif
   }
#ifdef SWAN
   sw_setnodeattr(GRAPHID(this), NODEID(i-1), GNODELABEL, " ");	  // Clear last
#endif
   num_vals--;
   return save;
}

//////////////////////////////////////////////////
bool Bucket::smaller()
//
// PURPOSE: Compares first two elements in bucket
// RETURNS: TRUE if the first element is less than the second, FALSE otherwise
{
   compcounter++;	   // Count the total number of comparisons
#ifdef SWAN
   memstm.seekp(0);  
   memstm << compcounter << ends;
   sw_setnodeattr(countg, NODEID(1), GNODELABEL, memstr);   // Show the count
   memstm.seekp(0);
   memstm << "Comparing " << bkt[0] << " and " << bkt[1] 
	  << ".  Press STEP to continue." << ends;
   sw_print(memstr);
   sw_setnodeattr(GRAPHID(this), NODEID(0), GNODEFILLED, TRUE);
   sw_setnodeattr(GRAPHID(this), NODEID(0), GNODECOLOR, LYELLOW);
   sw_setnodeattr(GRAPHID(this), NODEID(1), GNODEFILLED, TRUE);
   sw_setnodeattr(GRAPHID(this), NODEID(1), GNODECOLOR, LYELLOW);
   sw_displayallgraphs();	// I shouldn't need this before
   sw_break();			// all breaks!!
   sw_setnodeattr(GRAPHID(this), NODEID(0), GNODEFILLED, FALSE);
   sw_setnodeattr(GRAPHID(this), NODEID(0), GNODECOLOR, DGREEN);
   sw_setnodeattr(GRAPHID(this), NODEID(1), GNODEFILLED, FALSE);
   sw_setnodeattr(GRAPHID(this), NODEID(1), GNODECOLOR, DGREEN);
#endif
   return bkt[0] < bkt[1];
}

//////////////////////////////////////////////////
void Bucket::print()
//
// PURPOSE: Prints out contents of the bucket
{
   for (int i = 0; i < num_vals; i++) cout << bkt[i] << "  ";
   cout << endl;
}


// Global variables ----------------------------------------------------------
int Bucket::compcounter;			   // Automatically set to 0
Bucket topb(MAXLENGTH), lessb((MAXLENGTH+1)/2),		// The 4 buckets
       moreb((MAXLENGTH+1)/2), finalb(MAXLENGTH-2);

// Swan graph creation function ----------------------------------------------

#ifdef SWAN
//////////////////////////////////////////////////
void setup_swan(GS_MAIN_ARGS_F)
//
// PURPOSE: Creates and draws all of the graphs in the Swan window
{
   int i;
   topg = GRAPHID(&topb);	       // Graph IDs are addresses of buckets
   lessg = GRAPHID(&lessb);
   moreg = GRAPHID(&moreb);
   finalg = GRAPHID(&finalb);

   sw_init(GS_MAIN_ARGS_A);
   sw_disablebuttons(RUN | STEP) ;     // force viewer step through
   sw_step() ;			       // change mode to STEP

   make_swangraph(topg, BOX, DIRECTED, ARRAYACROSS);
   sw_setgraphattr(topg, GLAYOUTMODE, MANULAYOUT);
   sw_setgraphpos(topg, 60, 40);
   for (i = 0; i < MAXLENGTH; i++) sw_insertnode(topg, NODEID(i), " ");

   make_swangraph(lessg, BOX, DIRECTED, ARRAYACROSS);
   sw_setgraphattr(lessg, GLAYOUTMODE, MANULAYOUT);
   sw_setgraphpos(lessg, 0, 170);
   for (i = 0; i < (MAXLENGTH+1)/2; i++) sw_insertnode(lessg, NODEID(i), " ");

   make_swangraph(moreg, BOX, DIRECTED, ARRAYACROSS);
   sw_setgraphattr(moreg, GLAYOUTMODE, MANULAYOUT);
   sw_setgraphpos(moreg, 300, 170);
   for (i = 0; i < (MAXLENGTH+1)/2; i++) sw_insertnode(moreg, NODEID(i), " ");

   make_swangraph(finalg, BOX, DIRECTED, ARRAYACROSS);
   sw_setgraphattr(finalg, GLAYOUTMODE, MANULAYOUT);
   sw_setgraphpos(finalg, 85, 300);
   for (i = 0; i < MAXLENGTH-2; i++) sw_insertnode(finalg, NODEID(i), " ");

   make_swangraph(countg, BOX, DIRECTED, ARRAYACROSS);
   sw_setgraphattr(countg, GLAYOUTMODE, MANULAYOUT);
   sw_setgraphpos(countg, 480, 365);
   sw_insertnode(countg, NODEID(1), "");
   sw_setnodeattr(countg, NODEID(1), GNODEFILLED, TRUE);
   sw_setnodeattr(countg, NODEID(1), GNODECOLOR, PASGREEN);

   sw_displayallgraphs();
}
#endif

// Function definitions -----------------------------------------------------

//////////////////////////////////////////////////
void find_min_max()
//
// PURPOSE: This the is the routine that actually finds the minimum & maximum
//	    values.  This is what is being visualized with Swan
{
   while (topb.count() > 1) {	// Takes two each time, runs n/2 times
      if (topb.smaller())
	 { lessb.add(topb.remove()); moreb.add(topb.remove()); }
      else
	 { moreb.add(topb.remove()); lessb.add(topb.remove()); }
   }

   if (topb.count() == 1) {	// means n was odd--do one more comparison
#ifdef SWAN
      sw_print("Need to do one more comparison--take first number from "
	       "\"more\" bucket.");
      sw_displayallgraphs();	// Shouldn't need!
      sw_break();
#endif
      topb.add(moreb.remove());
      if (topb.smaller())
	 { lessb.add(topb.remove()); moreb.add(topb.remove()); }
      else
	 { moreb.add(topb.remove()); lessb.add(topb.remove()); }
   }

   // find smallest element in "less" bucket
   sw_print("Now, find min of the min's.");
   sw_displayallgraphs();
   sw_break();
   while (lessb.count() > 1) {
      if (lessb.smaller()) lessb.swap();
      finalb.add(lessb.remove());
   }

   // find largest element in "more" bucket
   sw_print("Now, find max of the max's");
   sw_displayallgraphs();
   sw_break();
   while (moreb.count() > 1) {
      if (!moreb.smaller()) moreb.swap();
      finalb.add(moreb.remove());
   }
}

//////////////////////////////////////////////////
int GS_MAIN(GS_MAIN_ARGS_F)
//
// PURPOSE: Driver for program
// PARAMETERS: The standard ones
{
   int num_to_sort = 0;
#ifdef SWAN
   setup_swan(GS_MAIN_ARGS_A);
   sw_print("How many numbers do you wish to use (15 max)?");
   sw_getstr(memstr);
   num_to_sort = atoi(memstr);
#else
 
   if (argc < 2) {
     cout << "How many numbers do you wish to use?  ";
     cin >> num_to_sort;
   }
   else num_to_sort = atoi(argv[1]);
#endif

   if (num_to_sort < 2 || num_to_sort > MAXLENGTH) {
      cerr << "ERROR: Use an integer between 2 and " << MAXLENGTH
	   << "." << endl;
      exit(EXIT_FAILURE);
   }

   // Fill array with random integers
   srand((unsigned) time(NULL));
   for(int i = 0; i < num_to_sort; i++)  {
      topb.add(rand() % MAXINT);
   }
  
#ifdef SWAN
   sw_enablebuttons(STEP);
   sw_print("Press STEP to begin.");
   sw_displayallgraphs();	// I shouldn't need this--bug with Swan
   sw_break();
#else
   // Print out array
   topb.print();
#endif

   // Now sort & print results
   find_min_max();

#ifdef SWAN
   sprintf(memstr, "Minimum value: %d   Maximum value: %d   Comparisons: %d",
           lessb.getelem(0), moreb.getelem(0),
		   finalb.comparison_count());
   sw_print(memstr);
   sw_displayallgraphs();	// I shouldn't need this--Swan bug!
   sw_break();
   sw_quit();
#else
   cout << "Minimum value is " << lessb.getelem(0)
        << ".\nMaximum value is " << moreb.getelem(0)
	<< ".\nNumber of comparions: " << finalb.comparison_count()
	<< endl; 
#endif
   return EXIT_SUCCESS;
}











